#!/usr/bin/python3
# -*- coding: utf-8 -*-

################################################################################
#
# Servidor Tablon - Cliente 1.5
# (c)Javier Fresno, 2015
#
# Servidor Tablon - Cliente 1.8
# (c) Arturo Gonzalez-Escribano, 2019
# Generalized for other contest types
#
################################################################################


# Imports
import sys
import binascii
import pickle
import re
import socket
from optparse import OptionParser
import getpass
import hashlib


# REPLACE SRP
from tablon._pysrp import *
# REPLACE SRP


# CLIENT VERSION
CLIENT_VERSION = '1.8'

DEFAULT_PORT = 80
OTHER_PORT = 8080


# MAX SIZE: This is checked again in the server
MAX_PROGRAM_SIZE = 15000 * 10 # This is 10 times the size of mpi version reference version (curso 1314)

#
# Class for the Secure session
#
class ClientSession():

	# Constructor
	def __init__(self):

		self.connected = False
		self.error = None

	# Connect and authenticate
	def connect(self,host,port,user,password):


		print ('Conecting ...')

		# 1. Connect the socket
		try:
			self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

			# Try to connect either to the given port or to other (80)
			try:
				self.socket.connect((host, port))
				print(('Connected to %s@%s:%d' % (user,host,port)))
			except:
				if not OTHER_PORT:
					raise
				port = OTHER_PORT
				self.socket.connect((host, port))
				print(('Connected to %s@%s:%d' % (user,host,port)))

			# DANIEL py2 -> py3: Chnage makefile mode to 'w' and 'r' and bufsize deleted
			self.wfile = self.socket.makefile('w')
			self.rfile = self.socket.makefile('r')

		except socket.error:
			self.error = "Error connecting to server"
			return

		# 2. Try to authenticate
		try:
			# Send start message
			#DANIEL py2 -> py3: Add flush() to send the message at the moment
			self.wfile.write('CONNECT /launcher HTTP/1.1\nHost: '+host+':'+str(port)+'\n\n')
			self.wfile.flush() #Flush it
			
			# Get server HTTP response
			ok = self.rfile.readline().strip()
			#print(ok)
			if ok[0:12] != 'HTTP/1.0 200':
				print( "Bad HTTP response when connecting" )
				raise Exception()
			# DANIEL py2 -> py3: Some data and empty line to close the HTTP response
			ok = self.rfile.readline().strip() #Server
			ok = self.rfile.readline().strip() #Date
			ok = self.rfile.readline().strip() #\r\n			

			# Send version
			self.wfile.write(CLIENT_VERSION + '\n')
			self.wfile.flush()
			
			# Get server ok
			ok = self.rfile.readline().strip()
			#print(ok)
			if ok[13:15] != 'OK':
				print(ok)
				sys.exit(-1)

			#Empty line to close the HTTP response			
			ok = self.rfile.readline().strip()

			print ('Authenticating user...')
			# Authenticate user
			usr = User(user, password, hash_alg=SHA256)
			uname, A = usr.start_authentication()
			# DANIEL py2 -> py3: Add decode() because wfile.write() only receives strings
			self.wfile.write(uname+'\n'+binascii.b2a_hex(A).decode()+'\n')
			self.wfile.flush()

			# Get the challenge
			s = binascii.a2b_hex(self.rfile.readline().strip())
			B = binascii.a2b_hex(self.rfile.readline().strip())
			
			# Process challenge and send response
			M = usr.process_challenge(s, B)
			if M is None:
				raise Exception()
			# DANIEL py2 -> py3: Add decode() because wfile.write() only receives strings
			self.wfile.write(binascii.b2a_hex(M).decode()+'\n')
			self.wfile.flush()			

			# Get the session
			HAMK = binascii.a2b_hex(self.rfile.readline().strip())
			usr.verify_session(HAMK)

			# Check authentication
			if usr.authenticated():
				print ("Successful authentication")
				self.connected = True
			else:
				raise Exception()

			# Get the session key
			self.key = usr.get_session_key()

		except:
			self.error = "Error authentication to server failed"
			return	

	# Send the request
	def send(self,request):

		# Msg
		print("Sending request")

		# Marshalling and send of the request
		data = pickle.dumps(request)
		self.wfile.write(str(len(data))+'\n')
		self.wfile.flush()
		# DANIEL py2 -> py3: data sent with socket.send() in order to send bytes object
		self.socket.send(data)	
	
		# Compute the hash of the file and our key
		m = hashlib.sha256()
		m.update(data)
		m.update(self.key)
		self.wfile.write(m.hexdigest()+'\n')
		self.wfile.flush()
		# Get the returned message
		self.socket.close()
		return self.rfile.read()

	# Close the socket
	def close(self):
		self.socket.close()






# Class that contains a request for the execution of a program
class ExecutionRequest():

	# Constructor
	def __init__(self,options):

		# Get the needed options
		self.nprocs = options.nprocs
		self.threads = options.threads
		self.program = base=os.path.basename(options.program)
		self.arguments = options.args
		self.queue = options.queue
		if options.reference != None:
			self.reference = True
			self.reference_name = options.reference
		else:
			self.reference = False
		self.timewall = options.timewall
		self.cl1 = options.cl1

		# Read the file
		# DANIEL py2 -> py3: file() doesn't exist in Python3, changed to open()
		with open(options.program) as f:
			self.content = f.read()

		# Copy kernels (if any)
		self.copy_kernels(options.program)
		
		if len(self.content) > MAX_PROGRAM_SIZE:
			print ("Program too big")
			sys.exit(-1)


	# Search for all cuda kernels
	def copy_kernels(self,programpath):

		regex = re.compile('^\s*[#]include\s*["](kernel.*[.]cu)["]',re.MULTILINE)
		includes = re.finditer(regex, self.content)

		# Iterate over the kernels
		for include in includes:
			self.copy_kernel(include.group(0),include.group(1),programpath)




	# Search for a cuda kernel and inserts it on the content
	def copy_kernel(self,kernelinclude,kernelname,programpath):

		print(("Including kernel %s" % kernelname))

		# Check if there is a kernel.cu in the same folder
		if os.path.dirname(programpath) == '':
			kfile = kernelname
		else:
			kfile = os.path.dirname(programpath) + '/' + kernelname
		if not os.path.exists(kfile):
			print(("Kernel %s not found" % kernelname))
			sys.exit()

		# Read the kernel
		# DANIEL py2 -> py3: file() doesn't exist in Python3, changed to open()
		with open(kfile) as f:
			kcontent = f.read()

		# Add comments
		kcontent = '//--------- KERNEL ---------\n' + kcontent + '//--------- KERNEL ---------\n'

		self.content = self.content.replace(kernelinclude,kcontent)







# Parse the input arguments
###############################################################################

parser = OptionParser(usage='usage: %prog [options] program.c/.cu/.cpp/.asm [program args]')

# Default
parser.set_defaults(port=DEFAULT_PORT)
parser.set_defaults(server='tfg-6506')
parser.set_defaults(nprocs=1)
parser.set_defaults(threads=1)
parser.set_defaults(reference=None)
parser.set_defaults(timewall=None)
parser.set_defaults(cl1=None)

# Group server
parser.add_option('-u','--user')
parser.add_option('-x','--password')
parser.add_option('-s','--server')
parser.add_option('-p','--port',type="int")
parser.add_option('-n','--nprocs',type="int")
parser.add_option('-t','--threads',type="int")
parser.add_option('-q','--queue')
parser.add_option('-r','--reference')
parser.add_option('-w','--timewall',type="int",help='Timewall in seconds')
parser.add_option('-c','--cl1',help='Flag para gestion cache/shared')

(options, args) = parser.parse_args()

if len(args) == 0:
	parser.print_help()
	sys.exit(-1)

if not options.user:
	print ("Missing user name")
	sys.exit(0)

if not options.password:
	options.password = getpass.getpass('Write your password: ')

if options.nprocs < 1:
	print ("Wrong number of processes")
	sys.exit(-1)

if options.threads < 1:
	print ("Wrong number of threads")
	sys.exit(-1)



options.program  = args[0]
options.args = ' '.join(args[1:])


if not options.program.endswith('.c') and not options.program.endswith('.cu') and not options.program.endswith('.cpp') and not options.program.endswith('.asm') and not options.program.endswith('.sonarqube'):
	parser.print_help()
	print ('Program must end with .c, .cpp, .cu, .asm or .sonarqube')
	sys.exit(-1)

# Check file
if not os.path.exists(options.program):
	parser.print_help()
	print ('Program do not exists')
	sys.exit(-1)



# Create the session
cs = ClientSession()
cs.connect(options.server,options.port,options.user,options.password)

# Exit if not connected
if not cs.connected:
	print((cs.error))
	sys.exit(-1)

# Create the request with the options
er = ExecutionRequest(options)

# Send the request and get the sever msg
msg = cs.send(er)
print (msg)


