#!/usr/bin/python3
# -*- coding: utf-8 -*-

################################################################################
#
# Servidor Tablon - LaunchServer
# Asignatura Computación paralela
# Javier Fresno, 2015
#
################################################################################


#
# Logging
#
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('tablon')

# Imports
import binascii, hashlib, pickle
from . import utils
from . import tablondb

#
# Import authentication library
#
from ._pysrp import *

#
# Instance to tablon module
#
import sys
tablon = sys.modules['tablon'] 




#
# The LaunchServer class is the handler that manages the client requests (CONNECT)
#
class LaunchServer:


	# CONNECT request is used to send data with the client
	def do_CONNECT(self):
		# We have a new connection
		logger.info("New connection from %s",self.client_address)
		
		# Authenticate the user
		if not self.authenticate():
			return

		# Manage the requst
		self.manage_request()

		return


	#
	# Authenticate the user
	#	
	def authenticate(self):

		# 1. Authentication
		try:
			# DANIEL py2 -> py3: Change self.wfile.write() for self.send_response() and 
			# self.send_response_only() to write correctly in the buffer
			# and add self.end_headers() after each sent message to flush them
			self.send_response(200, 'Connection Established')
			self.end_headers()
			# Get the version
			logger.info("Connection, reading client version")
			# DANIEL py2 -> py3: use decode() with the message read to get it as str
			client_version = self.rfile.readline().strip().decode()
			logger.info("Connection, client version %s",client_version)

			if(client_version != CONFIG.client_version):
				# DANIEL py2 -> py3: Use send_error to comunicate errors with the initial connection
				self.send_error(426,'Unsuported cliente version','Download the client again')
			else:
				# We send OK
				self.send_response_only(200, 'OK')
				self.end_headers()

			logger.info("Mandado el OK")

			# Get user name
			uname = self.rfile.readline().strip().decode()
			A = binascii.a2b_hex(self.rfile.readline().strip().decode())
			logger.debug("User %s connecting", uname)

			logger.info("Recibido usuario")

			# Get the user in our list
			user = tablon.users.get(uname)
			if not user:
				raise Exception('User do not exists')
			self.user = user

			# Create and send challenge
			svr = Verifier(uname, user.salt, user.vkey, A, hash_alg=SHA256)
			s, B  = svr.get_challenge()
			
			if s is None or B is None:
				raise Exception('Challenge failed')
			# DANIEL py2 -> py3: Change wfile.write() message to bytes
			self.wfile.write(b''.join([binascii.b2a_hex(s),b'\n']))
			self.wfile.write(b''.join([binascii.b2a_hex(B),b'\n']))
			logger.info("Challenge sent")

			# Get the solution to the challenge
			M = binascii.a2b_hex(self.rfile.readline().strip())
			logger.info("Solution recived")			

			# Verify session
			HAMK = svr.verify_session(M)

			if HAMK is None:
				raise Exception('Sesion not verified')

			# Send the session
			self.wfile.write(b''.join([binascii.b2a_hex(HAMK),b'\n']))
			if svr.authenticated():
				logger.info("User %s connected from %s", uname, self.client_address)
					
			# Get the session key
			key = svr.get_session_key()

			# Return
			self.username = uname
			self.userkey = key
			return True

		except Exception as e:
			user = tablondb.get_user_from_ip(self.client_address[0])
			logger.warning("Wrong connection from %s maybe '%s'", self.client_address, user)
			self.wfile.write(b'Wrong connection\n')

			print(e)
			import traceback
			print((traceback.format_exc()))


		return False
			

	#
	# Manage the request
	#	
	def manage_request(self):

		if not self.username:
			return

		try:
			# Get request len
			datalen = int(self.rfile.readline())
			if datalen > CONFIG.max_program_size:
				# DANIEL py2 -> py3: Change all self.wfile.write() to send bytes objects
				self.wfile.write(b'Request exceeds size limit')
				return
			# Check the restarttime
			if tablon.restartat != None:
				if (utils.get_timestamp() + CONFIG.time_before_restart) > tablon.restartat:
					self.wfile.write(b'Request not accepted, server is going to restart')
					return

			# Check quota v2
			if self.user.quota > 0 and not self.user.admin and self.user.quota * 1.25 < self.user._usedquota:
				if CONFIG.time_between_requests_no_quota_25 == 0:
					self.wfile.write(b'Quota exhausted' )
					return
				else: 
					now = utils.get_timestamp()
					last = tablondb.last_timestamp(self.username)

					if(last+CONFIG.time_between_requests_no_quota_25 > now):
						er = 'Quota exhausted (more than one quarter extra), only one request every %d seconds: Wait %d seconds more.' % (CONFIG.time_between_requests_no_quota_25,last+CONFIG.time_between_requests_no_quota_25-now)
						self.wfile.write(er.encode())
						return
			else:
				if self.user.quota > 0 and not self.user.admin and self.user.quota * 1.10 < self.user._usedquota:
					if CONFIG.time_between_requests_no_quota_10 == 0:
						self.wfile.write(b'Quota exhausted' )
						return
					else: 
						now = utils.get_timestamp()
						last = tablondb.last_timestamp(self.username)

						if(last+CONFIG.time_between_requests_no_quota_10 > now):
							er = 'Quota exhausted (more than one tenth extra), only one request every %d seconds: Wait %d seconds more.' % (CONFIG.time_between_requests_no_quota_10,last+CONFIG.time_between_requests_no_quota_10-now)
							self.wfile.write(er.encode())
							return
				else:
					if self.user.quota > 0 and not self.user.admin and self.user.quota < self.user._usedquota:
						if CONFIG.time_between_requests_no_quota == 0:
							self.wfile.write(b'Quota exhausted' )
							return
						else: 
							now = utils.get_timestamp()
							last = tablondb.last_timestamp(self.username)

							if(last+CONFIG.time_between_requests_no_quota > now):
								er = 'Quota exhausted (less than one tenth extra), only one request every %d seconds: Wait %d seconds more.' % (CONFIG.time_between_requests_no_quota,last+CONFIG.time_between_requests_no_quota-now)
								self.wfile.write(er.encode())
								return

			# Check the min time between requests
			if not self.user.admin:
				now = utils.get_timestamp()
				last = tablondb.last_timestamp(self.username)

				if(last+CONFIG.time_between_requests > now):
					er = 'Only one request every %d seconds' % CONFIG.time_between_requests
					self.wfile.write(er.encode())
					return

			# Read the data
			data = self.rfile.read(datalen)

			# Get the hash from the user
			usrhash = self.rfile.readline().strip().decode()

			# Compute the hash of the file and the key
			m = hashlib.sha256()
			m.update(data)
			m.update(self.userkey)
			
			# Check hash
			if usrhash != m.hexdigest():
				error = "Error at send, hash do not match"
				logger.warning(error + ", user %s", self.username)
				self.wfile.write(error.encode())
				return

			# Now its safe to read the data
			request = pickle.loads(data)
			
			# Sanitize the reques
			if not request.sanitize():
				self.wfile.write(b'Sanitize failed, wrong request, check compulsary parameters such as queue, etc.')
				return	

		except Exception as e:
			logger.warning("Error receiving request from %s: %s", self.username, self.client_address)
			self.wfile.write(b'Error receiving request\n')

			logger.warning(e)
			import traceback
			logger.warning(traceback.format_exc())

		try:
			# Remove the reference from not admins
			if not self.user.admin:
				request.reference = False

			# Everything ok at this point
			self.wfile.write(b'Request sent successfully\n')
			logger.info("Request received successfully from user %s %s",self.username,self.client_address)

			# Set the queue to empty if it is None
			if not request.queue:
				request.queue = ''

			# Check if the queue exists
			if not request.queue in tablon.queues:
				er = 'Queue "%s" do not exits' % request.queue
				self.wfile.write(er.encode())
				return

			# Check if the queue support the number of threads
			queue =  tablon.queues[request.queue]		
			ok, msg = queue.check_nprocs_threads(request)
			if not ok:
				self.wfile.write(msg.encode())
				return

			# Check extension
			ok, msg = queue.check_extension(request)
			if not ok:
				self.wfile.write(msg.encode())
				return

			# Check the leaderboard (if any)
			ok, msg = queue.check_leaderboard(request)
			if not ok:
				self.wfile.write(msg.encode())
				return

			# Check forbidden
			num,line = request.check_forbidden(queue.forbidden_code)
			if num:
				er = 'Forbidden code in Line %d: %s\n' % (num,line)
				self.wfile.write(er.encode())
				return

			request.user = self.username
			request.address = self.client_address[0]
			request.addressname = self.address_string()

			# Init the request and adds all the missing fields
			request.init()

			# Finally add the request and give it a rid
			rid = tablon.exemanager.add(request)

			# ARTURO: v7.3
			# Get the port of the web pages
			if hasattr(CONFIG, 'proxy_port'):
				port = CONFIG.proxy_port
			else:
				# Get the original port used to send the request
				if hasattr(CONFIG, 'real_port'):
					port = CONFIG.real_port
				else:
					try:
						port = utils.get_originalport(self.rfile._sock)
					except:
						port = CONFIG.port	

			# Check if port is 80 to skip the number in the URL
			if port == 80:
				str_port = ''
			else:
				str_port = ':%d' % 	CONFIG.port	

			# ARTURO v7.3: Add web_prefix	
			# Write the url 
			if hasattr(CONFIG, 'web_prefix'):
				url = "http://%s%s%srequest?rid=%d" % (CONFIG.host,CONFIG.web_prefix,str_port,rid)
			else:
				url = "http://%s%s/request?rid=%d" % (CONFIG.host,str_port,rid)
			er = 'Request id %d\n%s' % (rid,url)
			self.wfile.write(er.encode())

		except Exception as e:
			logger.warning("Error processing request from %s: %s", self.username, self.client_address)
			self.wfile.write(b'Error processing request. Contact the administrator to provide your request data.\n')

			logger.warning(e)
			import traceback
			logger.warning(traceback.format_exc())


