#!/usr/bin/python3
# -*- coding: utf-8 -*-

################################################################################
#
# Servidor Tablon - LeaderBoard
# Asignatura Computación paralela
# Javier Fresno, 2015
#
# v2
# Arturo Gonzalez-Escribano, 2020
#
################################################################################


#
# Logging
#
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('tablon')


#
# Instance to the tablon module
#
import sys
tablon = sys.modules['tablon'] 

# Imports
from .executionrequest import Status
import re
from . import tablondb
import os
import codecs


#
# Leaderboard Phase: a test the user program has to pass
#
class LeaderBoardPhase:

	# ARTURO: Add score field
	# ARTURO: Subphases
	def __init__(self,lb,override_request,ic_count,subphases,removeblanks_def,shouldpass_def,on_error_keep_scores_def):

		self.override_request = override_request
	
		self.lb = lb;
		self.subphases = subphases

		logger.info("Creating LeaderBoardItem")

		# ARTURO v2: Scores array
		self.scores = [0] * 10

		# ARTURO: Default values for subphases
		self.removeblanks_def = removeblanks_def
		self.shouldpass_def = shouldpass_def
		self.on_error_keep_scores_def = on_error_keep_scores_def

		# ARTURO: Add ic count or not
		self.ic_count = ic_count


	# Change the parameters and other attributes of the request object
	def override(self,request):
	
		logger.info("Antes de override")

		# ARTURO: Add the possibility of skipping arguments override (sonarqube feedback)
		if not self.override_request==None:
			for key in self.override_request:
				setattr(request, key, self.override_request[key])

		logger.info("Despues de override")


	#
	# ARTURO: Check errors or compare the results for every subphase/test
	#
	def check(self,request):

		# ARTURO v7.2: DETECT EXECUTION ERRORS, EXCEPTIONS, ETC.
		ok, msg = self.check_errors(request)

		# ARTURO v7.0: CHECK SUBPHASES
		cont_subphase=0
		cont_activated=0
		for subphase in self.subphases:
			cont_subphase += 1

			# ARTURO: Get subphase data
			self.results = subphase.get('results','')
			self.resulterror = subphase.get('resulterror','')
			self.exception = subphase.get('exception', False)
			self.shouldpass = subphase.get('shouldpass', self.shouldpass_def )
			self.removeblanks = subphase.get('removeblanks', self.removeblanks_def )
			self.on_error_keep_scores = subphase.get('on_error_keep_scores', self.on_error_keep_scores_def )
			self.regexp = subphase.get('regexp', False )
			# ARTURO v2: Get subphase scores array
			for i in range(10):
				self.scores[i] = subphase.get('score' + str(i), 0 )

			# ARTURO: Time limit for a task, should not be ignored by shoulpass value
			if not ok and msg == ": Time limit ":
				# ARTURO v2: Reset scores array
				# ARTURO v7.2: Optional keep scores
				if not self.on_error_keep_scores:
					request.scores = [0] * 10

				request.status = Status.ERROR
				request.message = "Error %s" % msg
				return False

			# ARTURO v2.1: No errors detected
			if ok:
				# ARTURO v2.1: Skip if exception subphases
				if self.exception:
					continue
					
				# ARTURO v2.1: Non exception subphases
				# ARTURO: Check results in stdout
				okResults, msg = self.check_results(request)

				# ARTURO: Debug
				if msg:
					logger.info("Req" + str(request.rid)+ ": check results: " + str(ok) + ", msg: " + msg )
				else:
					logger.info( "Req" + str(request.rid)+ ": check results: " + str(ok) + ", msg: None" )

				# ARTURO: Incorrect results
				if not okResults:	
 					# Force termination if this stage should pass
					if self.shouldpass:
						# ARTURO v2: Reset scores array
						# ARTURO v7.2: Optional keep scores
						if not self.on_error_keep_scores:
							request.scores = [0] * 10

						request.status = Status.ERROR
						request.message = "Error in result %s" % msg
						return False
					# Otherwise, skip scoring, go to next subphase
					else:
						continue

			# ARTURO v2.1: Errors detected
			else:
				# ARTURO v2.1: Non-exception subphase
				if not self.exception:
 					# Force termination if this stage should pass
					if self.shouldpass:
						# ARTURO v2: Reset scores array
						# ARTURO v7.2: Optional keep scores
						if not self.on_error_keep_scores:
							request.scores = [0] * 10
							
						request.status = Status.ERROR
						request.message = "Error in result %s" % msg
						return False
					# Otherwise, skip scoring, go to next subphase
					else:
						continue

				# ARTURO v2.1: Exception subphase, check type
				if not re.search(": " + self.results[0], msg):
					continue

			# ARTURO v2.1: Checks OK, add trigger scores
			cont_activated += 1
			for i in range( len(self.scores) ):
				request.scores[i] += self.scores[i]
		
	
		# ARTURO v7.2: All subphases pass or do not apply
		# ARTURO v7.2: If no error, process metric values
		if request.status != Status.ERROR :
			logger.info( "Trying to get scores from the results" )
			values = self.get_metrics_in_results(request,self.lb);

			# ARTURO v2: The "accumulate"  parameters indicate which scores should 
			#		be accumulated or substituted
			for i in range( len(values) ):
				if self.lb.metric_accumulate[i]:
					request.scores[i] += values[i]
				else:
					request.scores[i] = values[i]

		return True


	#
	# ARTURO: Check error strings in the output
	#
	def check_errors(self,request):

		# ARTURO v7.2: SEARCH SPECIFIC ERROR MESSAGES IN THE OUTPUT
		try:
			# ARTURO: CHECK EXECUTION ERRORS (Mars MIPS Simulator)
			if re.search('Error.*Runtime exception', request.output):
				return (False, ": Runtime exception")

			# ARTURO: CHECK BASIC EXECUTION ERRORS (1819)
			if re.search('-- Error in arguments:', request.output):
				return (False, ": Error in arguments")

			if re.search('-- Error in file:', request.output):
				return (False, ": Error in file")

			if re.search('-- Error allocating:', request.output):
				return (False, ": Error allocating")

			if re.search('Error allocating', request.output):
				return (False, ": Error allocating memory")

			# ARTURO: CHECK CUDA EXECUTION ERROR MESSAGES 
			if re.search('ErrorCUDA_malloc', request.output):
				return (False, ": CUDA Program reports: Error malloc")

			if re.search('ErrorCUDA_memcpy', request.output):
				return (False, ": CUDA Program reports: Error memcpy")

			if re.search('ErrorCUDA_kernel', request.output):
				return (False, ": CUDA Program reports: Error kernel")

			if re.search('ErrorCUDA_otro', request.output):
				return (False, ": CUDA Program reports: Error otro")

			# ARTURO: CHECK CUDA EXECUTION ERROR MESSAGES 1920
			if re.search('-- Error CUDA call:', request.output):
				return (False, ": CUDA Error in call check")

			if re.search('-- Error CUDA last:', request.output):
				return (False, ": CUDA Error in last-error check")

			# ARTURO: CHECK GENERAL C MEMORY MANAGEMENT PROBLEMS
			if re.search('Segmentation fault', request.output):
				return (False, ": Segmentation fault ")

			if re.search('double free or corruption', request.output):
				return (False, ": Memory management or access")

			if re.search('invalid next size', request.output):
				return (False, ": Memory management or access")

			if re.search('invalid pointer', request.output):
				return (False, ": Memory management or access")

			if re.search(' free', request.output):
				return (False, ": Memory management or access")

			if re.search('sysmalloc', request.output):
				return (False, ": Memory management or access")

			# ARTURO: CHECK PROCESS AND EXTERNAL QUEUE ERRORS
			if re.search('TIME LIMIT', request.output):
				return (False, ": Time limit ")

			if re.search('Killed', request.output):
				return (False, ": Task killed")

			if re.search('Node not available', request.output):
				return (False, ": Node not available, please notify administrator ")

		# If something is wrong, result is false with an error message
		# ARTURO: DEBUG VERSION
		except Exception as e:
			return (False, ": unexpected error :" + str(e))
		except:
			return (False, ": unexpected error")

		# ARTURO v7.2: ERROR STATUS WITH NO SPECIFIC ERROR MESSAGE IN OUTPUT
		if request.status == Status.ERROR:
			return (False, ": Process status error")
			
		return (True,None)


	#
	# ARTURO: Check results strings in the output
	#
	def check_results(self,request):

		try:
			# ARTURO: CHECK OUTPUT

			# ARTURO: LB items that look for exceptions do not need to check results
			if self.exception:
				return (False, ": Not a exception")


			# ARTURO: REMOVE BLANK SPACES
			if self.removeblanks:
				tocheck = re.sub(r"[ \t]+","", request.output )
			else:
				tocheck = request.output

			# ARTURO: SEARCH FOR REGULAR EXPRESSION IN THE OUTPUT
			if self.regexp:
				if re.search(self.results, request.output):
					return (True,None)
				else:
					return (False, ": result not valid")


			# ARTURO: CHECK OUTPUTS
			regex = re.compile('^Result:(.*)$',re.MULTILINE)
			values = re.finditer(regex, tocheck)

			# ITERATOR TO LIST
			values = list(values)

			# CHECK LEN
			if type(self.results) is not list:
				self.results = [self.results];		

			if len(values) != len(self.results):
				return (False, ": invalid number of results ")

			# CHECK EACH ELEMENT OF THE LIST
			for i, v in enumerate(values):
				value = v.group(1).strip()
				logger.info( "Req%s: Checking value %s, %s" % (request.rid,self.results[i],value) )

				if not self.check_result(i,value):
					return (False, ": result not valid")

		# If something is wrong, it is false
		except ValueError:
			return (False, ": value type error")
		# ARTURO: DEBUG VERSION
		except Exception as e:
			return (False, ": unexpected error :" + str(e))
		except:
			return (False, ": unexpected error")

		return (True,None)


	#
	# ARTURO: Check the value of one item in the results string
	#
	def check_result(self,i,value):

		if isinstance(self.results[i], int):
			value = int(value)
			return abs(value-self.results[i]) <= self.resulterror
		elif isinstance(self.results[i], float):
			value = float(value)
			return abs(value-self.results[i]) <= self.resulterror
		elif isinstance(self.results[i], str):
			return value == self.results[i]
		else:
			logger.warning( "Unexpected type in check_result %s, %s" % (self.results[i],value) )
			
		return False
	
		

	#
	# ARTURO: Get scores in results
	#
	def get_metrics_in_results(self,request,lb):

		logger.info(("Inside function get_metrics_in_results %s" % request.output ));

		try:
			# ARTURO v2: Score arrays in leaderboard
			names = lb.metric_capture

			# Search for the different Scores
			values = []
			# ARTURO v2: Ten scores with indexes 0-9
			#for i in range(9):
			for ind in range( len(names) ):
				search_str = ('%s: (.*[0-9\.]+)(.*)$' % names[ind] )
				regex = re.compile( search_str, re.MULTILINE )

				# ARTURO v2.1: Capture exception for malformed scores
				try:
					search = re.search(regex, request.output)
				except Exception as e:
					logger.info( "Score scanning, capture exception %s -- " % e )

				if search:
					# Get the value
					values.insert(ind, float( search.group(1) ))
					logger.info( ("Score %d captured from results: %f" % (ind,values[ind]) ) )
				else:
					values.insert(ind, float( 0 ))
					logger.info( ("Score %d not captured, reset to value %f" % (ind,values[ind]) ) )

			return values

		except Exception as e:
			logger.info( "Score capture, unexpected exception, stopping scores search %s -- " % e )
			pass
		
		return []


#
# A leaderboard
#
class LeaderBoard:

	# File to store the leaderboard
	FILE = 'serverdata/%s.lb'

	# Constructor
	def __init__(self,name,description='',isopen=True,**kwargs):

		logger.info("Creating leaderboard '%s'" % name)

		self.name = name
		self.description = description
		self.phases = []
		self.items = []
		self.isopen = isopen
		self.load()
		self.sort()
		self.save()
		# ARTURO v2: Score array makes many fields obsolete
		## ARTURO New fields for different classes of leaderboards
		self.classify = kwargs.get('classify','performance')
		# ARTURO New fields for default parameters on phases/subphases
		self.removeblanks_def = kwargs.get('removeblanks',False)
		self.shouldpass_def = kwargs.get('shouldpass',True)
		# ARTURO v7.2: Keep scores when a subphase does not pass
		self.on_error_keep_scores_def = kwargs.get('on_error_keep_scores',False)
		# ARTURO New field for total timewall
		self.total_timewall = kwargs.get('total_timewall',60)
		
		# ARTURO v2: Scores array
		self.metric_capture = kwargs.get('metric_capture', [] )
		self.metric_accumulate = kwargs.get('metric_accumulate', [] )
		self.metric_required = kwargs.get('metric_required', [] )
		self.score_title = kwargs.get('score_title', [] )
		self.score_formulae = kwargs.get('score_formulae', [] )
		self.score_format = kwargs.get('score_format', [] )
		self.score_minimum = kwargs.get('score_minimum', [] )
		self.score_classification = kwargs.get('score_classification', [] )

		if isopen:
			tablon.leaderboards[0:0] = [self]
		else:
			tablon.leaderboards.append(self)


	# ARTURO: Add phase with many possible outputs with different values
	def addPhase(self,override,subphases,ic_count=True):
		phase = LeaderBoardPhase( self, override, ic_count, subphases, self.removeblanks_def, self.shouldpass_def, self.on_error_keep_scores_def )
		self.phases.append(phase)


	def getPhase(self,idx):
		return self.phases[idx]

	def numPhases(self):
		return len(self.phases)



	#
	# Add a request into the leaderboard creating a LeaderBoardItem
	#
	def add(self,request):

		# Create and add a new item
		nitem = LeaderBoardItem(request)		
		self.items.append(nitem)
		tablondb.add_leaderboard(nitem)

		# sort
		self.sort()

		# save to file
		self.save()

	def sort(self):
		self.items = sorted(self.items, key=lambda i: (i.user, i.time))

	def save(self):

		directory = "serverdata"
		if not os.path.exists(directory):
			os.makedirs(directory)

		with codecs.open(self.FILE % self.name , 'w',"utf-8") as f:
			for item in self.items:
				f.write(item.toLine())

	def load(self):
		try:
			with codecs.open(self.FILE % self.name , 'r', "utf-8") as f:
				for line in f:
					nitem = LeaderBoardItem(line)
					self.items.append(nitem)
					# Remove completely
					if tablondb.initiated:
						tablondb.add_leaderboard(nitem)

		except :
			self.items = []


	# iterator: the array
	def __iter__(self):
		return iter(self.items)

	# len: the array
	def __len__(self):
		return len(self.items)



#
# An item of a leaderboard
#
class LeaderBoardItem:

	def __init__(self,request_or_line):

		if isinstance(request_or_line, str) or isinstance(request_or_line, str) :

			line = request_or_line.split(',')

			self.rid = int(line[0])
			self.user = line[1]
			self.program = line[2]
			self.queue = line[3]
			# ARTURO v2: Scores array
			self.scores = [0] * 10;
			self.scores[0] = line[4]
			self.scores[1] = line[5]
			self.scores[2] = line[6]
			self.scores[3] = line[7]
			self.scores[4] = line[8]
			self.scores[5] = line[9]
			self.scores[6] = line[10]
			self.scores[7] = line[11]
			self.scores[8] = line[12]
			self.scores[9] = line[13]
			self.rtime = float(line[14])
			self.timestamp = int(line[15])

		else:

			request = request_or_line

			self.rid = request.rid
			if request.reference:
				self.user = 'reference'
				self.program = request.reference_name
			else:
				self.user = request.user
				self.program = request.program
			self.queue = request.queue
			# ARTURO v2: Scores array
			#self.score = request.score
			self.scores = [0] * 10;
			for i in range(10):
				self.scores[i] = request.scores[i]
			#self.instructions = request.instructions
			self.time = request.lb_time
			self.rtime = request.time
			self.timestamp = request.timestamp


	def toLine(self):
		# ARTURO v2: Scores array
		return "%d,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%f,%d\n" % (self.rid,self.user,self.program,self.queue,self.scores[0],self.scores[1],self.scores[1],self.scores[1],self.scores[4],self.scores[5],self.scores[6],self.scores[7],self.scores[8],self.scores[9],self.time,self.rtime,self.timestamp)


