#!/usr/bin/python3
# -*- coding: utf-8 -*-

################################################################################
#
# Servidor Tablon - WebPages
# Asignatura Computación paralela
# Javier Fresno, 2015
#
# Asignaturas de Organización de Computadoras y Diseño de Software
# Arturo Gonzalez-Escribano 2019/2020
#
################################################################################


#
# Logging
#
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('tablon')

#
# Instance to the tablon module
#
import sys
tablon = sys.modules['tablon'] 

#
# Tablon imports
#
from .webserver2 import webrequest, web, render_template, Context, error_page
from . import tablondb

#
# Other imports
#
import re
from . import utils



@web.url('/ok')
def users_page():

	"""
	An OK json
	"""

	webrequest.content_type = 'application/json'

	return '{"isok":true,"version":"%s"}' % (tablondb.get_dbversion(),)




@web.url('/')
def root_page():

	"""
	The main page of the web
	"""
	
	context = Context()
	context['subtitle'] = CONFIG.web_title
	context['menuselected'] = 'Home'

	context['requests'] = tablondb.show_request_list()
	context['restartat'] = tablon.restartat

	return render_template('root.html',context)





@web.url('/users')
def users_page():

	"""
	The users page
	"""

	context = Context()
	context['title'] = 'Users'
	context['menuselected'] = 'Users'
	context['users'] = tablondb.show_users()

	return render_template('users.html',context)





@web.url('/faq',cache=60*60)
def faq_page():
	
	"""
	The fag page
	"""

	context = Context()
	context['title'] = 'FAQ'
	context['menuselected'] = 'FAQ'

	return render_template('faq.html',context)


@web.url('/doc',cache=60*60)
def doc_page():
	
	"""
	The doc page: Extra documentation for students, javadoc, etc.
	"""

	context = Context()
	context['title'] = 'Documentation'
	context['menuselected'] = 'doc'

	return render_template('doc.html',context)





@web.url('/client',cache=60*60)
def client_page():

	"""
	Get a the client script file
	"""

	# Change the content type
	webrequest.content_type = 'application/octet-stream'

	# Read the client script, add the pysrp library and other changes
	# DANIEL py2 -> py3: Changed 'rb' mode to 'r' to obtain str instead of bytes
	with open('client.py','r') as f:
		content = f.read()
	with open('tablon/_pysrp.py','r') as f:
		cont_pysrp = f.read()
	content = re.sub("'localhost'","'%s'" % CONFIG.host, content)

	if hasattr(CONFIG,"real_port"):
		content = re.sub("DEFAULT_PORT = 80","DEFAULT_PORT = %d" % CONFIG.real_port, content)
	else:
		content = re.sub("DEFAULT_PORT = 80","DEFAULT_PORT = %d" % CONFIG.port, content)

	# DANIEL v2.0: changed _pysrp import in client
	content = re.sub('from tablon[.]_pysrp import [*]', cont_pysrp, content)
	# DANIEL py2 -> py3: Delete all null bytes in the content result in order to work later
	content = content.encode().replace(b'\x00',b'').decode()
	
	# Compress if needed
	if CONFIG.compress_client:
		content = utils.gz_pack(content)

	# Send the client cache file
	return content




@web.url('/request')
def request_page():

	"""
	Page with the info of a particular request
	"""


	# Get the request id (if any)
	try:
		rid = int(webrequest.getParam('rid'))
	except:
		rid = None

	context = Context()
	context['title'] = 'No Request'
	if rid != None:
		rdb = tablondb.show_request(rid)
		if rdb:
			context['request'] = rdb			
			context['title'] = 'Request %05d' % rid

			# ARTURO v2: Calculate leaderboard info
			lb_active = rdb["leaderboard"]
			if ( lb_active != 0 ):
				queue_name = rdb["queue"]
				logger.info( "Request -> queue: " + queue_name )
				for lb in tablon.leaderboards:
					if ( lb.name == queue_name ):
						mylb = lb

				if ( mylb == None ):
					logger.error( "Request: Leaderboard not found: " + queue_name )

				leaderboard_values = [] 
				for i in range( len(mylb.score_title) ):
					try:
						leaderboard_values.append( { 'title':mylb.score_title[i] , 'value':eval( mylb.score_formulae[i], rdb ), 'minimum':mylb.score_minimum[i], 'format':mylb.score_format[i] } )
					except ZeroDivisionError as error:
						leaderboard_values.append( { 'title':mylb.score_title[i] , 'value':0.0, 'minimum':mylb.score_minimum[i], 'format':mylb.score_format[i] } )
						

				context['leaderboard_values'] = leaderboard_values

	return render_template('request.html',context)











@web.url('/tablon.css',ssl="Both")
def send_css():

	"""
	Tablon.css
	"""

	webrequest.content_type = 'text/css'
	webrequest.cachecontrol = True

	context = dict()
	context['color'] = CONFIG.color

	return render_template('tablon.css',context)





@web.url('/tablon.js',ssl="Both")
def send_js():

	"""
	Tablon.js 
	"""

	webrequest.content_type = 'application/javascript'
	webrequest.cachecontrol = True

	context = dict()

	return render_template('tablon.js',context)






@web.url('/leaderboards',cache=10)
def leaderboards_page():

	"""
	Page with all the leaderboards
	"""

	context = Context()
	context['title'] = 'Leaderboards'
	context['leaderboards'] = []
	context['menuselected'] = 'Leaderboards'

	for lb in tablon.leaderboards:
		items =  tablondb.show_leaderboard(lb)
		leaderboard = {'name':lb.name,'isopen':lb.isopen,'description':lb.description,'score_title':lb.score_title,'score_minimum':lb.score_minimum,'format':lb.score_format,'items':items}
		context['leaderboards'].append(leaderboard)

	return render_template('leaderboards.html',context)





@web.url('/queues',cache=60)
def queue_page():

	"""
	Page the with queues
	"""

	context = Context()
	context['title'] = 'Queues'
	context['menuselected'] = 'Queues'
	context['queues'] = tablon.onlyqueues

	return render_template('queues.html',context)



@web.url('/ajax_request_rows')
def ajax_request_rows():

	"""
	Ajax page to update the request page
	"""

	# Get the last request id (if any)
	try:
		rid = int(webrequest.getParam('rid'))
	except:
		rid = None

	# Send the html message
	context = dict()
	context['requests'] = tablondb.show_request_list(rid)

	return render_template('request_row.html',context)








